/*
 * Fillomino Solver in Copris
 * by Naoyuki Tamura
 * http://bach.istc.kobe-u.ac.jp/copris/puzzles/fillomino/
 */
package fillomino

import jp.kobe_u.copris._
import jp.kobe_u.copris.dsl._
import puzzle._

case class Fillomino(m: Int, n: Int, board: Seq[Seq[String]]) extends BoardPuzzle {
  def isBlank(cell: Cell) = at(cell) == "-"
  val maxRegionSize = math.max(20, numberCells.map(num(_)).max)
  def show(sol: Seq[Seq[Int]]) {
    for (i <- 0 until m) {
      for (j <- 0 until n)
	print("%3d".format(sol(i)(j)))
      println
    }
  }
}

object Solver extends BoardPuzzleSolver[Fillomino] {
  import scala.math.Ordering.Implicits._

  val name = "fillomino.Solver"
  var badPairOfRegions: Set[(Set[Cell],Set[Cell])] = _

  def puzzleFactory(m: Int, n: Int, board: Seq[Seq[String]]) =
    Fillomino(m, n, board)

  def define = {
    // Numbers
    for (cell <- puzzle.cells)
      if (puzzle.isNumber(cell))
	int('x(cell), puzzle.num(cell))
      else
	int('x(cell), 1, puzzle.maxRegionSize)
    // Construct each region as a spanning tree
    for (cell <- puzzle.cells; cell1 <- puzzle.adjCells(cell))
      int('e(cell,cell1), 0, 1)
    for (cell <- puzzle.cells; cell1 <- puzzle.adjCells(cell); if cell < cell1)
      add('e(cell,cell1) + 'e(cell1,cell) <= 1)
    for (cell <- puzzle.cells) {
      val in = puzzle.adjCells(cell).map(cell1 => 'e(cell1,cell))
      int('in(cell), 0, 1)
      add('in(cell) === Add(in))
    }
    // Each region consists of the same number cells
    for (cell <- puzzle.cells; cell1 <- puzzle.adjCells(cell); if cell < cell1)
      add((('e(cell,cell1) === 1) || ('e(cell1,cell) === 1)) ==>
	  ('x(cell) === 'x(cell1)))
    // Size of each region should be equal to the number
    for (cell <- puzzle.cells)
      int('s(cell), 1, puzzle.maxRegionSize)
    for (cell <- puzzle.cells) {
      val s = for (cell1 <- puzzle.adjCells(cell))
	      yield If('e(cell,cell1) === 0, Num(0), 's(cell1))
      add('s(cell) === Add(s) + 1)
      add(('in(cell) === 0) ==> ('s(cell) === 'x(cell)))
    }
  }

  def checkSolution: Boolean = {
    val nodes = puzzle.cells.toSet
    def adjs(cell: Cell) = puzzle.adjCells(cell).toSet.filter {
      cell1 => solution('e(cell,cell1)) > 0 || solution('e(cell1,cell)) > 0
    }
    val arcs = (for (cell <- puzzle.cells) yield cell -> adjs(cell)).toMap
    val components = getComponents(nodes, arcs)
    def bad(r1: Set[Cell], r2: Set[Cell]) =
      r1.exists(cell1 => puzzle.adjCells(cell1).exists(cell2 => r2.contains(cell2)))
    badPairOfRegions = for {
      (size,regions) <- components.groupBy(_.size).toSet
      Seq(r1,r2) <- regions.toSeq.combinations(2); if bad(r1, r2)
    } yield (r1,r2)
    if (verbose >= 2)
      println("badPairOfRegions = " + badPairOfRegions.size)
    badPairOfRegions.isEmpty
  }
  def addNegation {
    if (badPairOfRegions.isEmpty) {
      val cs = for (cell <- puzzle.cells; if ! puzzle.isNumber(cell))
	       yield 'x(cell) =/= solution('x(cell))
      add(Or(cs))
    } else {
      for ((r1,r2) <- badPairOfRegions) {
	val cs = for (cell <- r1 ++ r2; if ! puzzle.isNumber(cell))
		 yield 'x(cell) =/= solution('x(cell))
	add(Or(cs))
      }
    }
  }
  override def findFirstSolution = findIncremental(true, checkSolution, addNegation)
  override def findNextSolution = findIncremental(false, checkSolution, addNegation)

  def showSolution {
    val sol = for (i <- 0 until puzzle.m)
	      yield (0 until puzzle.n).map(j => solution('x((i,j))))
    if (quiet == 0) {
      println("Solution = " + sol)
      println("Size = " + sol.size)
      puzzle.show(sol)
    }
  }
}
