/*
 * Nurikabe Solver in Copris
 * by Naoyuki Tamura
 * http://bach.istc.kobe-u.ac.jp/copris/puzzles/nurikabe/
 */
package nurikabe

import jp.kobe_u.copris._
import jp.kobe_u.copris.dsl._
import puzzle._

case class Nurikabe(m: Int, n: Int, board: Seq[Seq[String]]) extends BoardPuzzle {
  def isBlank(cell: Cell) = at(cell) == "-"
  val numOfBlackCells = m * n - cells.filter(isNumber(_)).map(num(_)).sum
  def show(sol: Set[Cell]) {
    for (i <- 0 until m) {
      for (j <- 0 until n) {
	val cell = (i,j)
	if (isNumber(cell)) print("%2d ".format(num(cell)))
	else if (sol.contains(cell)) print("## ")
	else print(".. ")
      }
      println
    }
    println
  }
}

object Solver extends BoardPuzzleSolver[Nurikabe] {
  import scala.math.Ordering.Implicits._

  val name = "nurikabe.Solver"
  var blackComponents: Set[Set[Cell]] = _

  def puzzleFactory(m: Int, n: Int, board: Seq[Seq[String]]) =
    Nurikabe(m, n, board)

  def define = {
    // Regions ('x(i,j): region index, 0 means black cell)
    val regionMax = puzzle.numberCells.size
    var regionIndex = 0
    for (cell <- puzzle.cells) {
      if (puzzle.isNumber(cell)) {
	regionIndex += 1
	int('x(cell), regionIndex)
      } else {
	int('x(cell), 0, regionMax)
      }
    }
    // Black cells do not make 2x2 blocks
    for ((i,j) <- puzzle.cells; if i < puzzle.m - 1 && j < puzzle.n - 1)
      add(('x((i,j)) > 0) || ('x((i+1,j)) > 0) || ('x((i,j+1)) > 0) || ('x((i+1,j+1)) > 0))
    // Different white regions are not connected
    for ((i,j) <- puzzle.cells) {
      if (i < puzzle.m - 1)
	add((('x((i,j)) > 0) && ('x((i+1,j)) > 0)) ==> ('x((i,j)) === 'x((i+1,j))))
      if (j < puzzle.n - 1)
	add((('x((i,j)) > 0) && ('x((i,j+1)) > 0)) ==> ('x((i,j)) === 'x((i,j+1))))
    }
    // White region has specified size
    // Construct a tree for each white region
    for (cell <- puzzle.cells; cell1 <- puzzle.adjCells(cell)) {
      int('e(cell,cell1), 0, 1)
      add(('e(cell,cell1) === 1) ==> (('x(cell) > 0) && ('x(cell1) > 0)))
    }
    for (cell <- puzzle.cells; cell1 <- puzzle.adjCells(cell); if cell < cell1)
      add('e(cell,cell1) + 'e(cell1,cell) <= 1)
    for (cell <- puzzle.cells) {
      val in = puzzle.adjCells(cell).map(cell1 => 'e(cell1,cell))
      int('in(cell), 0, 1)
      add('in(cell) === Add(in))
      if (puzzle.isNumber(cell))
	add('in(cell) === 0)
      else
	add(('x(cell) > 0) ==> ('in(cell) === 1))
    }
    // Count the size of each white region
    val maxSize = puzzle.numberCells.map(cell => puzzle.num(cell)).max
    for (cell <- puzzle.cells)
      int('s(cell), 1, maxSize)
    for (cell <- puzzle.cells) {
      val s = for (cell1 <- puzzle.adjCells(cell))
	      yield If('e(cell,cell1) === 0, Num(0), 's(cell1))
      add('s(cell) === Add(s) + 1)
      if (puzzle.isNumber(cell))
	add('s(cell) === puzzle.num(cell))
    }
    // Hints
    for (cell <- puzzle.numberCells
	 if puzzle.num(cell) == 1; cell1 <- puzzle.adjCells(cell))
      add('x(cell1) === 0)
    for (cell <- puzzle.numberCells;
	 dij <- Seq((1,0), (0,1)); cell1 = puzzle.move(puzzle.move(cell, dij), dij)
	 if puzzle.isCell(cell1) && puzzle.isNumber(cell1))
      add('x(puzzle.move(cell, dij)) === 0)
    for (cell <- puzzle.numberCells;
	 dj <- Seq(-1, 1); cell1 = puzzle.move(cell, (1,dj))
	 if puzzle.isCell(cell1) && puzzle.isNumber(cell1)) {
      add('x(puzzle.move(cell, (1,0))) === 0)
      add('x(puzzle.move(cell, (0,dj))) === 0)
    }
    if (puzzle.numOfBlackCells >= 2) {
      for (cell <- puzzle.cells; if puzzle.isBlank(cell))
	add(('x(cell) === 0) ==> Or(puzzle.adjCells(cell).map('x(_) === 0)))
    }
  }

  def checkSolution: Boolean = {
    val nodes = puzzle.cells.filter(cell => solution('x(cell)) == 0).toSet
    def nextCells(cell: Cell) =
      puzzle.adjCells(cell).filter(cell1 => solution('x(cell1)) == 0).toSet
    val arcs = nodes.map(cell => cell -> nextCells(cell)).toMap
    blackComponents = getComponents(nodes, arcs)
    if (verbose >= 2)
      println("Components = " + blackComponents.size)
    blackComponents.size == 1
  }
  def addNegation {
    for (blacks <- blackComponents) {
      val whites = for (cell <- blacks; cell1 <- puzzle.adjCells(cell); if solution('x(cell1)) > 0)
		   yield cell1
      add(Or(blacks.map(cell => 'x(cell) > 0)) || Or(whites.map(cell => 'x(cell) === 0)))
    }
  }
  override def findFirstSolution = findIncremental(true, checkSolution, addNegation)
  override def findNextSolution = findIncremental(false, checkSolution, addNegation)

  def showSolution {
    require(blackComponents.size == 1)
    val sol = blackComponents.head
    if (quiet == 0) {
      println("Solution = " + sol)
      println("Size = " + sol.size)
      puzzle.show(sol)
    }
  }
}
