/*
 * Polarium Solver in Copris
 * by Naoyuki Tamura
 * http://bach.istc.kobe-u.ac.jp/copris/puzzles/polarium/
 */
package polarium

import jp.kobe_u.copris._
import jp.kobe_u.copris.dsl._
import puzzle._

case class Polarium(m: Int, n: Int, board: Seq[Seq[String]]) extends BoardPuzzle {
  def isBorder(cell: Cell) = at(cell) == "+"
  def isBlack(cell: Cell) = at(cell) == "*"
  def isWhite(cell: Cell) = at(cell) == "."
  val rows = 1 until m-1
  def blacks(i: Int) = (0 until n).count(j => isBlack((i,j)))
  def whites(i: Int) = (0 until n).count(j => isWhite((i,j)))
  val minFlips = {
    val flips = for (i <- rows) yield math.min(blacks(i), whites(i))
    flips.sum
  }
  val maxFlips = {
    val flips = for (i <- rows) yield math.max(blacks(i), whites(i))
    val nodes = for (i <- 0 until m; j <- 0 until n) yield (i,j)
    flips.sum + nodes.count(isBorder(_))
  }
  def show(sol: Seq[Cell]) {
    val arcs = sol.sliding(2).map(e => (e(0),e(1))).toSet
    def connected(cell1: Cell, cell2: Cell) =
      arcs.contains((cell1,cell2)) || arcs.contains((cell2,cell1))
    for (i <- 0 until m) {
      for (j <- 0 until n) {
	val cell = (i,j)
	print(at(cell))
	if (j+1 < n && connected(cell, (i,j+1))) print ("--")
	else print ("  ")
      }
      println
      for (j <- 0 until n) {
	val cell = (i,j)
	if (i+1 < m && connected(cell, (i+1,j))) print ("|  ")
	else print ("   ")
      }
      println
    }
  }
}

object Solver extends BoardPuzzleSolver[Polarium] {
  import scala.math.Ordering.Implicits._

  val name = "polarium.Solver"

  def puzzleFactory(m: Int, n: Int, board: Seq[Seq[String]]) =
    Polarium(m, n, board)

  def root = (puzzle.m,puzzle.n)
  def cells = root +: puzzle.cells
  def adjCells(cell: Cell) =
    if (cell == root) puzzle.cells
    else root +: puzzle.adjCells(cell)

  def define {
    for (i <- puzzle.rows)
      int('flip(i), 0, 1)
    for (cell <- cells; cell1 <- adjCells(cell))
      int('e(cell,cell1), 0, 1)
    for (cell <- cells; cell1 <- adjCells(cell); if cell < cell1)
      add('e(cell,cell1) + 'e(cell1,cell) <= 1)
    for (cell <- cells) {
      val in = adjCells(cell).map(cell1 => 'e(cell1,cell))
      val ot = adjCells(cell).map(cell1 => 'e(cell,cell1))
      int('io(cell), 0, 1)
      add('io(cell) === Add(in))
      add('io(cell) === Add(ot))
      if (cell == root)
	add('io(cell) === 1)
      else if (puzzle.isBlack(cell))
	add('flip(cell._1) === 'io(cell))
      else if (puzzle.isWhite(cell))
	add('flip(cell._1) =/= 'io(cell))
    }
    for (cell <- cells)
      int('x(cell), 0, puzzle.maxFlips)
    for (cell <- cells; cell1 <- adjCells(cell); if cell1 != root)
      add(('e(cell,cell1) > 0) ==> ('x(cell) < 'x(cell1)))
    if (options.contains("opt")) {
      int('flips, puzzle.minFlips, puzzle.maxFlips)
      minimize('flips)
      for (cell <- cells)
	add('x(cell) <= 'flips)
      // Hints
      val s = for (i <- puzzle.rows)
	yield 'flip(i) * (puzzle.blacks(i) - puzzle.whites(i)) + puzzle.whites(i)
      add(Add(s) <= 'flips)
    }
  }

  def showSolution {
    def path(cell: Cell): Seq[Cell] = {
      val cell1 = adjCells(cell).find(cell1 => solution('e(cell,cell1)) > 0).get
      if (cell1 == root) Seq.empty
      else cell1 +: path(cell1)
    }
    val sol = path(root)
    if (quiet == 0) {
      println("Solution = " + sol)
      println("Size = " + sol.size)
      puzzle.show(sol)
    }
  }
}
